from __future__ import absolute_import
input_name = '../examples/linear_elasticity/material_nonlinearity.py'
output_name = 'test_material_nonlinearity.vtk'

from tests_basic import TestInput
class Test(TestInput):
    pass
