# -*- Mode: Python -*-
"""
Basic types.
"""
cimport numpy as np
import numpy as np

ctypedef np.complex128_t complex128
ctypedef np.float64_t float64
ctypedef np.int32_t int32
ctypedef np.uint32_t uint32
