from __future__ import absolute_import
input_name = '../examples/linear_elasticity/linear_viscoelastic.py'
output_name_trunk = 'test_linear_viscoelastic'

from tests_basic import TestInputEvolutionary

class Test(TestInputEvolutionary):
    pass
