from __future__ import absolute_import
input_name = '../examples/multi_physics/thermo_elasticity.py'
output_name = 'test_thermo_elasticity.vtk'

from tests_basic import TestInput
class Test(TestInput):
    pass
