from __future__ import absolute_import
input_name = '../examples/large_deformation/perfusion_tl.py'
output_name_trunk = 'test_perfusion_tl'

from tests_basic import TestInputEvolutionary

class Test(TestInputEvolutionary):
    pass
