from __future__ import absolute_import
input_name = '../examples/dg/advection_diffusion_2D.py'
output_name = 'advection_diffusion_sol.msh'

from tests_basic import TestInput
class Test( TestInput ):
    pass